import pytest
from google.cloud.firestore_v1 import DocumentSnapshot, DocumentReference

from onto.view.query_delta import make_snapshot
from .fixtures import CTX
from .test_domain_model import setup_cities, City, StandardCity

from onto.query.cmp import v


def test_where():
    from onto.query.query import Query
    from onto.domain_model import DomainModel
    from onto import attrs

    class Foo(DomainModel):
        s = attrs.bproperty()

    q = Foo.get_query().where(v.s==1)
    from onto.mapper.fields import argument
    assert q.arguments == [argument(key='s', comparator='==', val=1)]
    q = q._to_firestore_query()
    assert q is not None


@pytest.mark.usefixtures("setup_cities")
def test_query_with_attr_str():

    expected_dict = {
        'Washington D.C.': {
            'cityName': 'Washington D.C.',
            'country': 'USA',
            'capital': True,
            'obj_type': "Municipality",
            'doc_id': 'DC',
            'doc_ref': 'City/DC'
        },
        'San Francisco': {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
            'obj_type': "StandardCity",
            'doc_id': 'SF',
            'doc_ref': 'City/SF'
        },
        'Los Angeles': {
            'cityName': 'Los Angeles',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'socal'],
            'obj_type': "StandardCity",
            'doc_id': 'LA',
            'doc_ref': 'City/LA'
        }

    }

    res_dict = dict()

    for obj in City.where(country=("==", "USA")):
        d = obj.to_dict()
        res_dict[d["cityName"]] = d

    assert res_dict['Washington D.C.'] == expected_dict['Washington D.C.']
    assert res_dict['San Francisco'] == expected_dict['San Francisco']
    assert res_dict['Los Angeles'] == expected_dict['Los Angeles']


@pytest.mark.usefixtures("setup_cities")
def test_query_with_cmp():

    expected_dict = {
        'Washington D.C.': {
            'cityName': 'Washington D.C.',
            'country': 'USA',
            'capital': True,
            'obj_type': "Municipality",
            'doc_id': 'DC',
            'doc_ref': 'City/DC'
        },
        'San Francisco': {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
            'obj_type': "StandardCity",
            'doc_id': 'SF',
            'doc_ref': 'City/SF'
        },
        'Los Angeles': {
            'cityName': 'Los Angeles',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'socal'],
            'obj_type': "StandardCity",
            'doc_id': 'LA',
            'doc_ref': 'City/LA'
        }

    }

    res_dict = dict()


    for obj in City.where(City.country == "USA", StandardCity.city_state == "CA"):
        d = obj.to_dict()
        res_dict[d["cityName"]] = d

    assert res_dict['San Francisco'] == expected_dict['San Francisco']
    assert res_dict['Los Angeles'] == expected_dict['Los Angeles']


def test_trigger_snapshot(CTX):
    data = {"oldValue": {}, "updateMask": {},
            "value": {"createTime": "2020-06-03T00:23:18.348623Z",
                      "fields": {"a": {"stringValue": "b"}},
                      "name": "projects/flask-boiler-testing/databases/(default)/documents/gcfTest/36ea7LTtYJHpW4yJQCp2",
                      "updateTime": "2020-06-03T00:23:18.348623Z"}}

    snapshot = make_snapshot(data['value'], client=CTX.dbs.firestore_client)
    assert isinstance(snapshot, DocumentSnapshot)
    assert snapshot.to_dict() == {'a': 'b'}
    assert snapshot.create_time is not None
    assert snapshot.update_time is not None

    assert make_snapshot(dict(), client=CTX.dbs.firestore.firestore_client) is None



from onto.attrs import attribute

from onto.models.base import Serializable


class H(Serializable):

    i = attribute.Attribute(type_cls=int, doc="I am 'i'.")

    j = attribute.PropertyAttribute(
        type_cls=str,
        doc="""I am 'j'. See next line. 
    This is my second line. 
    """
    )


from collections import namedtuple

graph_schema = namedtuple('graph_schema', ['op_type', 'name', 'graphql_object_type'])


def test__schema_cls_from_attributed_class():
    # import asyncio
    # loop = asyncio.get_event_loop()

    from onto.models.utils import _graphql_object_type_from_attributed_class
    attributed = H

    graphql_schema = _graphql_object_type_from_attributed_class(attributed)

    async def sub(parent, info, **kwargs):
        pass
    #
    from graphql import GraphQLObjectType
    query_schema = GraphQLObjectType(
        name='Query',
        fields={
            'h': graphql_schema
        }
    )

    from gql import query, subscribe

    @query
    async def h(parent, info, **kwargs):
        return {
            'i': 1,
            'j': 'one'
        }

    graph_schema(op_type='Query', name='h', graphql_object_type=graphql_schema)

    @subscribe
    async def h(parent, info, **kwargs):
        # Register topic
        # Listen to topic 
        for i in range(5):
            import asyncio
            await asyncio.sleep(i)
            yield {
                'h': {
                    'i': i,
                    'j': f"number is {i}"
                }

            }


    subscription_schema = GraphQLObjectType(
        name='Subscription',
        fields={
            'h': graphql_schema
        }
    )

    from graphql import GraphQLSchema
    schema = GraphQLSchema(
        query=query_schema,
        subscription=subscription_schema
    )

    from stargql import GraphQL

    async def on_startup():
        from asyncio.queues import Queue
        global q
        q = Queue()

    async def shutdown():
        pass

    app = GraphQL(
        schema=schema,
        on_startup=[on_startup],
        on_shutdown=[shutdown]
    )

    # import uvicorn
    # uvicorn.run(app, port=8080, debug=True)
    # return app



# import functools
#
# from onto.common import _NA
# from onto.database import Database, Reference, Snapshot, Listener
# from onto.context import Context as CTX
#
# import kafka
#
# # TODO: NOTE maximum of 1 firestore client allowed since we used a global var.
# from typing import List
# from onto.query.query import Query
# from onto.store.snapshot_container import SnapshotContainer
# from math import inf
#
# FirestoreListener = None
#
#
# class KafkaReference(Reference):
#
#     def is_collection(self):
#         return len(self.params) % 2 == 1
#
#     @property
#     def collection(self):
#         return self.first
#
#     def is_document(self):
#         return len(self.params) % 2 == 0
#
#     @classmethod
#     def from__document_name(cls, document_name: str):
#         return cls.from_str(document_name)
#
#     @property
#     def _document_path(self):
#         return str(self)
#
#
# class KafkaReadDatabase(Database):
#
#     def __init__(self, collection_id: str):
#         self.collection_id = collection_id
#         self.consumer = AIOKafkaConsumer(
#             'my_topic', 'my_other_topic',
#             bootstrap_servers='localhost:9092',
#             group_id="my-group")
#
#     async def start_consumer(self):
#         consumer =
#         # Get cluster layout and join group `my-group`
#         await consumer.start()
#         try:
#             # Consume messages
#             async for msg in consumer:
#                 print("consumed: ", msg.topic, msg.partition, msg.offset,
#                       msg.key, msg.value, msg.timestamp)
#         finally:
#             # Will leave consumer group; perform autocommit if enabled.
#             await consumer.stop()
#
#     async def ensure_timestamp(self, ts):
#         """
#         A helper method to wait for last_poll_timestamp to exceed the timestamp.
#         Used to make sure that all messages are loaded are server is ready.
#         """
#
#
#     @classmethod
#     def listener(cls):
#         from onto.database.utils import GenericListener
#         return GenericListener
#
#     bootstrap_servers = None
#
#     d = dict()
#
#     @classmethod
#     def _onto_set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
#         cls.d[str(ref)] = snapshot.to_dict()
#         cls.listener()._pub(reference=ref, snapshot=snapshot)
#
#     @classmethod
#     def get(cls, ref: Reference, transaction=_NA):
#         return Snapshot(cls.d[str(ref)])
#
#     update = set
#     create = set
#
#     @classmethod
#     def delete(cls, ref: Reference, transaction=_NA):
#         """ Note: this only deletes one instance that has _doc_id == ref.last
#
#         :param ref:
#         :param transaction:
#         :return:
#         """
#         del cls.d[str(ref)]
#
#
# class KafkaSnapshot(Snapshot):
#
#     @classmethod
#     def from_data_and_meta(
#             cls, **kwargs):
#         """
#
#         :param kwargs:
#         :return:
#         """
#         DATA_KEYWORD = 'data'
#         if DATA_KEYWORD not in kwargs:
#             raise ValueError
#         else:
#             data = kwargs[DATA_KEYWORD]
#             __onto_meta__ = {
#                 key: val
#                 for key, val in kwargs.items()
#                 if key != DATA_KEYWORD
#             }
#             return cls(**data, __onto_meta__=__onto_meta__)
#
#     @classmethod
#     def empty(cls, **kwargs):
#         return cls.from_data_and_meta(
#             data=dict(),
#             exists=False,
#             **kwargs
#         )
#



from collections import UserDict, namedtuple
from onto.store.snapshot_container import SnapshotContainer

struct_ref = namedtuple(
    'struct_ref', ['dm_cls', 'id', 'ref', 'snapshot', 'obj'],
    defaults=[None, None, None, None, None])


class Struct(UserDict):

    def __init__(self, schema_obj=None, schema_cls=None):
        super().__init__()
        if schema_obj is None:
            schema_obj = schema_cls()
        self.schema_obj = schema_obj

    @property
    def vals(self):
        for _, val in self.data.items():
            if isinstance(val, dict):
                for _, v in val.items():
                    yield v
            else:
                yield val

    def get_item(self, key):
        """ Only return when key is defined. Used to access values of
        this object.

        :param key:
        :return:
        """
        if key not in self.data.keys():
            return None
        else:
            return self.data[key]

    def __getitem__(self, key):
        """
        Initializes a field to dict if it was not declared before

        :param item:
        :return:
        """

        if key not in self.data.keys():
            self.data[key] = dict()
        return super().__getitem__(key)


class SnapshotStruct(UserDict):

    def __init__(self, schema_obj=None, schema_cls=None):
        super().__init__()
        if schema_obj is None:
            schema_obj = schema_cls()
        self.schema_obj = schema_obj

    @property
    def vals(self):
        for _, val in self.data.items():
            if isinstance(val, dict):
                for _, v in val.items():
                    yield v
            else:
                yield val

    def __getitem__(self, key):
        """
        Initializes a field to dict if it was not declared before
        Used to write to this object.

        :param item:
        :return:
        """

        if key not in self.data.keys():
            self.data[key] = dict()
        return super().__getitem__(key)

    def to_struct_and_container(self):
        container = SnapshotContainer()
        struct = Struct(schema_obj=self.schema_obj)
        for key, val in self.data.items():
            if isinstance(val, dict):
                for k, v in val.items():
                    _cls, snapshot = v
                    struct[key][k] = (_cls, snapshot.reference.id)
                    container.set(snapshot.reference._document_path, snapshot)
            else:
                _cls, snapshot = val
                struct[key] = (_cls, snapshot.reference.id)
                container.set(snapshot.reference._document_path, snapshot)
        return struct, container



from onto import schema, fields


def test_create():

    class ModelASchema(schema.Schema):
        int_a = fields.Integer(load_from="intA", dump_to="intA")

    class ModelAASchema(ModelASchema):
        int_aa = fields.Integer(load_from="intAA", dump_to="intAA")

    from onto.models.factory import ClsFactory
    ModelAA = ClsFactory.create(
        name="ModelAA",
        schema=ModelAASchema
    )

    obj_aa = ModelAA.new()
    assert hasattr(obj_aa, "int_a")
    assert hasattr(obj_aa, "int_aa")


def test_create_with_new():

    class ModelACSchema(schema.Schema):
        int_a = fields.Integer(load_from="intA", dump_to="intA")

    from onto.models.factory import ClsFactory
    ModelAC = ClsFactory.create_customized(
        name="ModelAC",
        schema=ModelACSchema,
        importable=False,
        exportable=True
    )

    obj_a = ModelAC.new(int_a=3)

    assert obj_a.int_a == 3


def test_create_with_new_exportable():

    class ModelACSchema(schema.Schema):
        int_a = fields.Integer(load_from="intA", dump_to="intA")

    from onto.models.factory import ClsFactory
    ModelAC = ClsFactory.create_customized(
        name="ModelAC",
        schema=ModelACSchema,
        importable=False,
        exportable=True
    )

    obj_a = ModelAC.new(int_a=3)

    assert set(obj_a.to_dict().items()) >= set({
        "intA": 3
    }.items())




import asyncio
from datetime import datetime
# from graphene import ObjectType, String, Schema, Field
from aiokafka import AIOKafkaConsumer
from stargql.subscription import Subscription

from onto.source.base import Source

"""
[Start] Sample code from aiokafka 
ref: https://github.com/aio-libs/aiokafka 
"""
loop = asyncio.get_event_loop()
async def _kafka_subscribe(topic_name):
    consumer = AIOKafkaConsumer(
        topic_name,
        bootstrap_servers='localhost:9092',
        # group_id="my-group"
    )
    # Get cluster layout and join group `my-group`
    await consumer.start()
    # TODO: make first yield when listener has started
    try:
        # Consume messages
        async for msg in consumer:
            yield msg
            # print("consumed: ", msg.topic, msg.partition, msg.offset,
            #       msg.key, msg.value, msg.timestamp)
    finally:
        # Will leave consumer group; perform autocommit if enabled.
        await consumer.stop()
loop.run_until_complete(_kafka_subscribe(topic_name='my-topic'))


"""
[End] Sample code from aiokafka 
Ref: https://github.com/aio-libs/aiokafka
"""

# # Every schema requires a query.
# class Query(ObjectType):
#     hello = String()
#
#     def resolve_hello(root, info):
#         return "Hello, world!"
#
#
# class Subscription(ObjectType):
#     time_of_day = String()
#
#     async def subscribe_time_of_day(root, info):
#         while True:
#             yield datetime.now().isoformat()
#             await asyncio.sleep(1)
# schema = Schema(query=Query, subscription=Subscription)

async def main(schema):
    subscription = 'subscription { timeOfDay }'
    result = await schema.subscribe(subscription)
    async for item in result:
        print(item.data['timeOfDay'])

# asyncio.run(main(schema))

#
# def _as_graphql_root_schema(attributed):
#     from graphql import GraphQLSchema
#
#     from onto.models.utils import _graphql_object_type_from_attributed_class
#     graphql_ot = _graphql_object_type_from_attributed_class(attributed)
#
#     async def sub(parent, info, **kwargs):
#         pass
#
#     schema = GraphQLSchema(
#         query=graphql_ot,
#         subscription=Subscription
#     )
#
#     from graphql.subscription import create_source_event_stream
#     create_source_event_stream(schema=schema, document=schema.ast_node, )
#     _ = subscribe(schema=schema, subscribe_field_resolver=sub)
#
#     return schema


class GraphQLSource(Source):

    def __init__(self, attributed_cls):
        super().__init__()
        self.attributed_cls = attributed_cls

    def start(self):
        schema = _as_graphql_root_schema(self.attributed_cls)
        from stargql import GraphQL

        async def on_startup():
            from asyncio.queues import Queue
            global q
            q = Queue()

        async def shutdown():
            pass

        app = GraphQL(
            schema=schema,
            on_startup=[on_startup],
            on_shutdown=[shutdown]
        )
        return app


#
# class KafkaSource(Source):
#
#     # def __init__(self, query):
#     #     """ Initializes a ViewMediator to declare protocols that
#     #             are called when the results of a query change. Note that
#     #             mediator.start must be called later.
#     #
#     #     :param query: a listener will be attached to this query
#     #     """
#     #     super().__init__()
#     #     self.query = query
#
#     def start(self):
#         self._register()
#
#     def _register(self):
#         res = _kafka_subscribe(self.topic_name, self._call)
#         asyncio.run(res)
#         async for item in res:
#
#     @classmethod
#     def delta(cls, container):
#         start = container._read_times[-2]
#         end = container._read_times[-1]
#         for key in container.d.keys():
#             for snapshot in container.get_with_range(key, start, end):
#                 from onto.database import Snapshot
#                 prev: Snapshot = snapshot.prev
#                 cur: Snapshot = snapshot
#                 if not prev.exists:
#                     yield ("on_create", key, cur)
#                 elif prev.exists and cur.exists:
#                     yield ("on_update", key, cur)
#                 elif prev.exists and not cur.exists:
#                     yield ("on_delete", key, cur)
#                 else:
#                     raise ValueError
#
#     def _call(self, msg):
#         self._invoke_mediator(
#             func_name=func_name, ref=ref, snapshot=snapshot)



from collections import namedtuple
#
# def bproperty(*args, **kwargs):
#     from .attribute import PropertyAttribute
#     from pony.orm.core import Required
#     class C(Required, PropertyAttribute):
#         pass
#         # def __init__(self, *args, **kwargs):
#         #     super().__init__(str, *args, **kwargs)
#     return C(str, *args, **kwargs)


from .attribute import AttributeBase, \
    RelationshipAttribute as relation, \
    PropertyAttribute as bproperty, \
    DictAttribute as bdict, \
    EmbeddedAttribute as embed, \
    ObjectTypeAttribute as object_type, \
    LocalTimeAttribute as local_time, \
    StringAttribute as string, \
    PropertyAttribute as integer, \
    DocRefAttribute as doc_ref, \
    PropertyAttribute as action

from .attribute_new import AttributeBase
with AttributeBase._get_root() as attrs:
    attrs = attrs

# TODO: change import for string and integer



from typing import Type

from onto.domain_model import DomainModel
from onto.source.base import Source


class MockDomainModelSource(Source):

    def __init__(self, dm_cls: Type[DomainModel]):
        super().__init__()
        self.dm_cls: Type[DomainModel] = dm_cls
        self.thread = None

    async def _invoke_mediator(self, *, func_name, ref, snapshot):
        obj = self.dm_cls.from_snapshot(ref=ref, snapshot=snapshot)
        try:
            await super()._invoke_mediator_async(func_name=func_name, obj=obj)
        except Exception as _:
            import logging
            logging.exception(f'async _invoke_mediator failed for {func_name} {str(ref)}')

    def start(self, loop=None):
        if loop is None:
            import asyncio
            loop = asyncio.get_event_loop()
        self._register(loop=loop)

    import asyncio

    def _get_awaitable(self):
        return self.dm_cls._datastore().listener().listen(col=self.dm_cls._get_collection_name(), source=self)

    def _register(self, loop: asyncio.BaseEventLoop):
        from onto.context import Context as CTX
        _awaitable = self._get_awaitable()
        _ = loop.create_task(_awaitable)


    def get_all(self):
        return self.dm_cls._datastore().listener().get_all(col=self.dm_cls._get_collection_name())



import pytest


def test_hello():
    from onto.helpers import make_variable

    ct1 = make_variable('ct1', default='ct1default')
    ct2 = make_variable('ct2', default='ct2default')

    with ct1('ct1a'):
        assert ct1.get() == 'ct1a'
        with ct1('ct1b'):
            assert ct1.get() == 'ct1b'

    assert ct1.get() == 'ct1default'



import pytest

from examples.meeting_room.domain_models import Meeting, User
from examples.meeting_room.view_models.meeting_session import MeetingSession
from onto.database import Snapshot
from onto.view.mutation import PatchMutation
from onto.query.query import ViewModelQuery
from onto.view.document import \
    ViewMediatorDAV
from onto import testing_utils
# Import the fixtures used by fixtures
from onto.context import Context
from tests.fixtures import CTX
from .fixtures import *
from examples.meeting_room.view_models.user_view import UserViewDAV


def test_start_flask(users, tickets, location, meeting, CTX):

    from examples.meeting_room.views.rest_mediators import MeetingSessionRest
    from flask import Flask
    app = Flask(__name__)
    MeetingSessionRest.start(app)

    testing_utils._wait()

    test_client = app.test_client()

    res = test_client.get(
        path='meeting_sessions/meeting_1')

    assert res.json.items() >= {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   # 'doc_ref': 'users/tijuana/MeetingSessionDAV/meeting_1',
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}.items()

    res = test_client.get(
        path='meeting_sessions/')
    assert res == dict()


def test_start(users, tickets, location, meeting, CTX):

    from examples.meeting_room.views.mediators import MeetingSessionShow
    MeetingSessionShow.start()

    testing_utils._wait()

    ref = Context.db.ref / "users" / users[0].doc_id / MeetingSession.__name__ / meeting.doc_id
    assert CTX.dbs.leancloud.get(ref).to_dict().items() >= {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   # 'doc_ref': 'users/tijuana/MeetingSessionDAV/meeting_1',
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}.items()

    CTX.db.delete(ref=ref)


def test_start_fs(users, tickets, location, meeting, CTX):

    from examples.meeting_room.views.mediators import MeetingSessionGet
    MeetingSessionGet.start()

    testing_utils._wait()

    ref = Context.db.ref / "users" / users[0].doc_id / MeetingSession.__name__ / meeting.doc_id
    assert CTX.db.get(ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   # 'doc_ref': 'users/tijuana/MeetingSessionDAV/meeting_1',
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    CTX.db.delete(ref=ref)



@pytest.fixture
def delete_after(request, CTX):
    def fin():
        patch_collection_name = "{}_PATCH".format(MeetingSession.__name__)
        col_ref = Context.db.ref/'**'/patch_collection_name
        q = ViewModelQuery(ref=col_ref)
        for ref, _ in CTX.db.query(q):
            CTX.db.delete(ref=ref)
    request.addfinalizer(finalizer=fin)


def test_mutate(users, tickets, location, meeting, delete_after, CTX):
    """
    Tests that mutate actions affect domain model and view model.

    :param users:
    :param tickets:
    :param location:
    :param meeting:
    :return:
    """
    # mediator = MeetingSessionGet()
    # mediator.start()

    from examples.meeting_room.views.mediators import MeetingSessionGet
    from examples.meeting_room.views.mediators import MeetingSessionPatch
    MeetingSessionGet.start()
    MeetingSessionPatch.start()  # TODO: isolate state between tests

    testing_utils._wait(factor=.7)

    # testing_utils._wait(60)  # TODO: delete; used right now for breakpoint

    """
    Tests that MeetingSessionGet works 
    """
    ref = Context.db.ref/"users"/users[0].doc_id/MeetingSession.__name__/meeting.doc_id
    assert CTX.db.get(ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   # 'doc_ref': 'users/tijuana/MeetingSessionDAV/meeting_1',
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    # testing_utils._wait(500)

    """
    Tests that MeetingSessionPatch works 
    """
    patch_ref = Context.db.ref/'users'/'tijuana'/'MeetingSession_PATCH'/meeting.doc_id
    CTX.db.update(
        ref=patch_ref,
        snapshot=Snapshot(
            target_meeting_id="meeting_1",
            inSession=False
        )
    )

    testing_utils._wait(factor=.7)

    m = Meeting.get(doc_id="meeting_1")
    assert m.status == "closed"
    #
    # ref = Context.db.collection("users").document(users[0].doc_id) \
    #     .collection(MeetingSession.__name__).document(meeting.doc_id)
    assert CTX.db.get(ref=ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   'inSession': False,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    for user in users:
        reference = CTX.db.ref/"users"/user.doc_id/MeetingSession.__name__/meeting.doc_id
        CTX.db.delete(reference)


def test_mutate_fs(users, tickets, location, meeting, delete_after, CTX):
    """
    Tests that mutate actions affect domain model and view model.

    :param users:
    :param tickets:
    :param location:
    :param meeting:
    :return:
    """
    # mediator = MeetingSessionGet()
    # mediator.start()

    from examples.meeting_room.views.mediators import MeetingSessionGet
    from examples.meeting_room.views.mediators import MeetingSessionPatch
    MeetingSessionGet.start()
    MeetingSessionPatch.start()  # TODO: isolate state between tests

    testing_utils._wait(factor=.7)

    # testing_utils._wait(60)  # TODO: delete; used right now for breakpoint

    """
    Tests that MeetingSessionGet works 
    """
    ref = Context.db.ref/"users"/users[0].doc_id/MeetingSession.__name__/meeting.doc_id
    assert CTX.db.get(ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   # 'doc_ref': 'users/tijuana/MeetingSessionDAV/meeting_1',
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    # testing_utils._wait(500)

    """
    Tests that MeetingSessionPatch works 
    """
    patch_ref = Context.db.ref/'users'/'tijuana'/'MeetingSession_PATCH'/meeting.doc_id
    CTX.db.update(
        ref=patch_ref,
        snapshot=Snapshot(
            target_meeting_id="meeting_1",
            inSession=False
        )
    )

    testing_utils._wait(factor=.7)

    m = Meeting.get(doc_id="meeting_1")
    assert m.status == "closed"
    #
    # ref = Context.db.collection("users").document(users[0].doc_id) \
    #     .collection(MeetingSession.__name__).document(meeting.doc_id)
    assert CTX.db.get(ref=ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   'inSession': False,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    for user in users:
        reference = CTX.db.ref/"users"/user.doc_id/MeetingSession.__name__/meeting.doc_id
        CTX.db.delete(reference)


@pytest.fixture
def delete_meeting_sessions(request, CTX):
    def fin():
        col_ref = Context.db.ref/'**'/MeetingSession.__name__
        q = ViewModelQuery(ref=col_ref)
        for ref, _ in CTX.db.query(q):
            CTX.db.delete(ref=ref)
    request.addfinalizer(finalizer=fin)


def test_domain_model_changes(users, tickets, location, meeting, delete_meeting_sessions, CTX):
    """ Tests that view model updates when domain model is changed
    The feature is currently broken; it requires an active connection for all
        models in store

    :param users:
    :param tickets:
    :param location:
    :param meeting:
    :return:
    """
    from examples.meeting_room.views.mediators import MeetingSessionGet
    mediator = MeetingSessionGet()

    mediator.start()

    testing_utils._wait(factor=2)

    ref = Context.db.ref/"users"/users[0].doc_id/MeetingSession.__name__/meeting.doc_id

    assert CTX.db.get(ref=ref).to_dict() == {'latitude': 32.880361,
                                   'numHearingAidRequested': 2,
                                   'attending': [
                                       {'hearing_aid_requested': True,
                                        'name': 'Joshua Pendergrast',
                                        'organization': 'SDSU'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': False,
                                        'name': 'Thomasina Manes'},
                                       {'organization': 'UCSD',
                                        'hearing_aid_requested': True,
                                        'name': 'Tijuana Furlong'}],
                                   'longitude': -117.242929,
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    # User Tijuana Furlong sets their attendance to False
    tickets[0].attendance = False
    tickets[0].save()

    testing_utils._wait()

    """
    Expect the document to be updated to exclude Tijuana Furlong from a
      list of people attending the meeting session and the hearing aid
      counter.
    """
    assert CTX.db.get(ref=ref).to_dict()  == {'latitude': 32.880361,
                                   'numHearingAidRequested': 1,
                                   'attending': [
                                       {
                                           'name': 'Joshua Pendergrast',
                                           'organization': 'SDSU',
                                           'hearing_aid_requested': True},
                                       {
                                           'name': 'Thomasina Manes',
                                           'organization': 'UCSD',
                                           'hearing_aid_requested': False}
                                   ],
                                   'longitude': -117.242929,
                                   'inSession': True,
                                   'address': '9500 Gilman Drive, La Jolla, CA'}

    for user in users:
        Context.db.collection("users").document(user.doc_id) \
            .collection(MeetingSession.__name__).document(meeting.doc_id) \
            .delete()


def test_view_model(users, tickets, location, meeting, delete_meeting_sessions):
    meeting_session = MeetingSession.get(doc_id=meeting.doc_id,
                                         once=True)

    assert meeting_session.to_dict() == \
           {'inSession': True,
            'longitude': -117.242929,
            'latitude': 32.880361,
            'address': '9500 Gilman Drive, La Jolla, CA',
            'attending': [
                {
                    'name': 'Joshua Pendergrast',
                    'organization': 'SDSU',
                    'hearing_aid_requested': True},
                {
                    'name': 'Thomasina Manes',
                    'organization': 'UCSD',
                    'hearing_aid_requested': False},
                {
                    'name': 'Tijuana Furlong',
                    'organization': 'UCSD',
                    'hearing_aid_requested': True},
            ],
            'numHearingAidRequested': 2}


class UserViewMediatorDAV(ViewMediatorDAV):

    def __init__(self, *args, user_cls=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_cls = user_cls

    @classmethod
    def notify(cls, obj):
        doc_ref = CTX.db.ref/"UserViewDAV"/obj.user_id
        doc_ref.set(obj.to_dict())

    # def generate_entries(self):
    #     d = dict()
    #     users = self.user_cls.all()
    #     for user in users:
    #         assert isinstance(user, User)
    #         user_id = user.doc_id
    #         obj = self.view_model_cls.get(
    #             user_id=user_id,
    #             once=False,
    #             f_notify=self.notify
    #         )
    #         d[Context.db.document(f"UserViewDAV/{user_id}")._document_path] = obj
    #     return d


def test_user_view(users, tickets, location, meeting):
    user_view = UserViewDAV.get(user_id="thomasina", )
    user_view.wait_for_first_success()
    # time.sleep(3)  # TODO: delete after implementing sync

    assert user_view.to_dict() == {
        'firstName': 'Thomasina',
        'hearingAidRequested': False,
        'lastName': 'Manes',
        'meetings': [{'address': '9500 Gilman Drive, La Jolla, CA',
                      'attending': [{'hearing_aid_requested': True,
                                     'name': 'Joshua Pendergrast',
                                     'organization': 'SDSU'},
                                    {'hearing_aid_requested': False,
                                     'name': 'Thomasina Manes',
                                     'organization': 'UCSD'},
                                    {'hearing_aid_requested': True,
                                     'name': 'Tijuana Furlong',
                                     'organization': 'UCSD'}],
                      'inSession': True,
                      'latitude': 32.880361,
                      'longitude': -117.242929,
                      'numHearingAidRequested': 2}],
        'obj_type': 'UserViewDAV',
        'organization': 'UCSD'}


@pytest.mark.skip
def test_user_view_diff(users, tickets, location, meeting):
    user_view = UserViewDAV.get(user_id="thomasina", )

    # time.sleep(3)  # TODO: delete after implementing sync

    updated_dict = user_view._export_as_view_dict().copy()
    updated_dict["lastName"] = 'M.'

    res = user_view.diff(
        updated_dict
    )
    assert res == {'lastName': 'M.'}


def test_propagate_change(users, tickets, location, meeting, CTX):
    user_id = users[1].doc_id

    user_view = UserViewDAV.get(user_id=user_id, )
    user_view.wait_for_first_success()

    # time.sleep(3)

    user_view.store.user.last_name = "M."
    user_view.propagate_change()

    user_ref = Context.db.ref/"users"/user_id
    assert CTX.db.get(ref=user_ref).to_dict()["lastName"] == "M."


class UserViewMutationDAV(PatchMutation):
    pass


def test_mutation(users, tickets, location, meeting):
    user_id = users[1].doc_id

    mediator = UserViewMediatorDAV(
        view_model_cls=UserViewDAV,
        user_cls=User,
        mutation_cls=UserViewMutationDAV
    )

    mediator.start()

    ref = Context.db.collection("UserViewDAV").document(user_id)

    # time.sleep(3)  # TODO: delete after implementing sync
    #
    user_view = mediator.instances[Context.db.document(f"UserViewDAV/{user_id}")._document_path]

    ref.collection("_PATCH_UserViewDAV").add({
        "lastName": "Manes-Kennedy"
    })

    testing_utils._wait(factor=.7)

    updated_user = User.get(doc_id=user_id)

    assert updated_user.last_name == "Manes-Kennedy"

    res = user_view.to_dict()

    assert res['lastName'] == "Manes-Kennedy"



# from onto.store.snapshot_container import SnapshotContainer
# from onto.store.struct import Struct
# from .color_fixtures import Color, PaletteDomainModel
# from onto import fields
# from onto.store.business_property_store import BusinessPropertyStore, BPSchema
# import pytest
#
#
# class SomeSchema(BPSchema):
#     favorite_color = fields.StructuralRef(dm_cls=Color)
#
#
# @pytest.mark.usefixtures("CTX")
# def test_get_manifest(color_refs):
#
#     cian_id = color_refs[0].id
#
#     struct = Struct(schema_obj=SomeSchema())
#     struct["favorite_color"] = (Color, cian_id)
#
#     g, gr, manifest = BusinessPropertyStore._get_manifests(struct, SomeSchema())
#
#     assert g == {
#         "favorite_color": 'projects/flask-boiler-testing/databases/(default)/documents/colors/doc_id_cian'
#     }
#     assert gr == {
#         'projects/flask-boiler-testing/databases/(default)/documents/colors/doc_id_cian': ["favorite_color"]
#     }
#     assert manifest == {
#         (Color, 'projects/flask-boiler-testing/databases/(default)/documents/colors/doc_id_cian'),
#     }
#
#
# def test_update(CTX, color_refs):
#     cian_id = color_refs[0].id
#     struct = Struct(schema_obj=SomeSchema())
#     struct["favorite_color"] = (Color, cian_id)
#
#     class Store(BusinessPropertyStore):
#         pass
#
#     store = Store(struct=struct, snapshot_container=SnapshotContainer())
#     store._container.set(
#         'projects/flask-boiler-testing/databases/(default)/documents/colors/doc_id_cian',
#             CTX.db.document('projects/flask-boiler-testing/databases/(default)/documents/colors/doc_id_cian').get()
#     )
#     store.refresh()
#     assert isinstance(store.favorite_color, Color)
#
#
# def test_obj_options(CTX, color_refs, request):
#     palette = PaletteDomainModel.new(
#         doc_id="partial_rainbow_palette",
#         palette_name="partial_rainbow",
#         colors=color_refs.copy()
#     )
#     palette.save()
#
#     def fin():
#         palette.delete()
#
#     request.addfinalizer(fin)
#
#     class PaletteStoreSchema(BPSchema):
#         favorite_palette = fields.StructuralRef(dm_cls=PaletteDomainModel)
#
#     struct = Struct(schema_obj=PaletteStoreSchema())
#     struct["favorite_palette"] = (PaletteDomainModel, palette.doc_id)
#
#     class Store(BusinessPropertyStore):
#         pass
#
#     """
#     Original: non-nested retrieves DocumentReference
#     """
#
#     store = Store(
#         struct=struct,
#         snapshot_container=SnapshotContainer()
#     )
#     store._container.set(
#         'projects/flask-boiler-testing/databases/(default)/documents/PaletteDomainModel/partial_rainbow_palette',
#             CTX.db.document('projects/flask-boiler-testing/databases/(default)/documents/PaletteDomainModel/partial_rainbow_palette').get()
#     )
#     store.refresh()
#     assert isinstance(store.favorite_palette, PaletteDomainModel)
#     assert isinstance(store.favorite_palette.colors[0], Color)
#
#     # """
#     # New: nested retrieves objects
#     # """
#     # store = Store(
#     #     struct=struct,
#     #     snapshot_container=SnapshotContainer(),
#     #     obj_options=dict(
#     #         must_get=True
#     #     )
#     # )
#     # store._container.set(
#     #     'projects/flask-boiler-testing/databases/(default)/documents/PaletteDomainModel/partial_rainbow_palette',
#     #     CTX.db.document(
#     #         'projects/flask-boiler-testing/databases/(default)/documents/PaletteDomainModel/partial_rainbow_palette').get()
#     # )
#     # store.refresh()
#     # assert isinstance(store.favorite_palette, PaletteDomainModel)
#     # assert isinstance(store.favorite_palette.colors[0], Color)



import pytest as pytest
from google.type.color_pb2 import Color

from onto import schema, fields, domain_model, view_model, \
    attrs
from onto.store import reference
from onto.store.store import BPSchema, Store
from onto.registry import ModelRegistry
from onto.store.struct import Struct
from onto.view_model import ViewModel
from onto.models import factory

class ColorSchema(schema.Schema):
    name = fields.Str()


class ColorDomainModelBase(domain_model.DomainModel):
    _collection_name = "colors"


class RainbowStoreBpss(BPSchema):
    colors = fields.StructuralRef(dm_cls=Color, many=True)


@pytest.fixture
def rainbow_vm(CTX):

    if ModelRegistry.get_cls_from_name("RainbowViewModelDAV") is not None:
        return ModelRegistry.get_cls_from_name("RainbowViewModelDAV")

    class RainbowStore(Store):
        colors = reference(dm_cls=Color, many=True, missing=list)


    class RainbowViewModelDAV(ViewModel):

        rainbow_name = attrs.bproperty(import_enabled=False)
        colors = attrs.bproperty(import_enabled=False)

        doc_ref = attrs.bproperty(export_enabled=False)

        # @doc_ref.getter
        # def doc_ref(self):
        #     return self._doc_ref

        obj_type = attrs.bproperty(import_enabled=False, export_enabled=False)

        @colors.getter
        def colors(self):
            return [self.store.colors[key].name for key in sorted(self.store.colors)]

        @rainbow_name.getter
        def rainbow_name(self):
            return "-".join(self.colors)

        @classmethod
        def get(cls, color_names: str=None, once=True, **kwargs):
            color_name_list = color_names.split("+")
            struct = dict()
            struct["colors"] = {
                "doc_id_{}".format(color_name):
                    (Color, "doc_id_{}".format(color_name))
                for color_name in color_name_list
            }
            store = RainbowStore.from_struct(struct=struct)
            vm_id = color_names
            doc_ref = CTX.db.ref/"RainbowDAV"/vm_id
            return super().get(doc_ref=doc_ref,
                               store=store,
                               once=once,
                               **kwargs)

        def save(self):
            CTX.db.set(
                ref=self.doc_ref,
                snapshot=self.to_snapshot()
            )

    return RainbowViewModelDAV


@pytest.fixture
def color_refs(request):

    cian = Color.new("doc_id_cian")
    cian.name = 'cian'
    cian.save()

    magenta = Color.new("doc_id_magenta")
    magenta.name = "magenta"
    magenta.save()

    yellow = Color.new("doc_id_yellow")
    yellow.name = "yellow"
    yellow.save()

    black = Color.new("black")
    black.name = "black"
    black.save()

    def fin():
        cian.delete()
        magenta.delete()
        yellow.delete()
        black.delete()

    request.addfinalizer(fin)

    return [cian.doc_ref,
            magenta.doc_ref,
            yellow.doc_ref,
            black.doc_ref
            ]


class Palette(schema.Schema):
    palette_name = fields.Str()
    colors = fields.Relationship(nested=True, many=True)


class PaletteViewModelBase(view_model.ViewModel):
    pass


@pytest.fixture
def vm(color_refs, CTX, request):

    # Create palette document in firestore
    vm = PaletteViewModel.new(
        doc_ref=CTX.db.collection("palettes").document("palette_id_a")
    )
    vm.palette_name = 'cmyk'
    vm.colors = color_refs

    vm.save()

    def fin():
        vm.delete()

    request.addfinalizer(fin)

    return vm


Color = factory.ClsFactory.create(
    name="Color",
    schema=ColorSchema,
    base=ColorDomainModelBase
)


class ColorViewModel(view_model.ViewModel):

    color = attrs.relation(export_enabled=False, nested=True, dm_cls=Color)
    name = attrs.bproperty()

    @name.getter
    def name(self):
        return self.color.name


class PaletteViewModel(view_model.ViewModel):
    palette_name = attrs.bproperty()
    colors = attrs.embed(obj_cls=ColorViewModel, many=True)
    doc_ref = attrs.doc_ref()

    @colors.getter
    def colors(self):
        return [
            ColorViewModel.new(color=Color.get(doc_ref=color_ref))
            for color_ref in self._attrs.colors
        ]


PaletteDomainModel = factory.ClsFactory.create(
    name="PaletteDomainModel",
    schema=Palette,
    base=domain_model.DomainModel
)



from collections import UserDict, UserList

import pytest

from onto import fields, attrs
from onto import schema as fb_schema
from marshmallow import fields as marshmallow_fields, schema

from unittest import mock

from onto.domain_model import DomainModel
from onto.firestore_object import FirestoreObject, \
    _nest_relationship_import
from onto.mapper.helpers import RelationshipReference
from .fixtures import CTX


@pytest.fixture
def mfunc(monkeypatch):
    F = marshmallow_fields.Field
    mfunc = mock.MagicMock(return_value=None)
    monkeypatch.setattr(F, "__init__", mfunc)
    return mfunc


def test_init_read_only():

    int_a_field = marshmallow_fields.Integer(dump_only=True)

    class TrivialSchema(fb_schema.Schema):

        int_a = int_a_field

    assert int_a_field.dump_only


def test_relationship_many(CTX):

    doc_ref_1, doc_ref_2 = \
        CTX.db.ref/"hello/1", CTX.db.ref/"hello/2"

    class ContainsIterableSchema(schema.Schema):
        the_iterable = marshmallow_fields.Dict(
            values=fields.Relationship(nested=False, obj_type=None))

    fd = ContainsIterableSchema().fields["the_iterable"]

    res = fd.deserialize(value={1: doc_ref_1, 2: doc_ref_2})

    assert res == {1: RelationshipReference(nested=False, doc_ref=doc_ref_1),
                   2: RelationshipReference(nested=False, doc_ref=doc_ref_2)}

    class ContainsIterableSchema(schema.Schema):
        the_iterable = fields.Relationship(nested=False, many=True, obj_type=None)

    fd = ContainsIterableSchema().fields["the_iterable"]

    res = fd.deserialize(value=[doc_ref_1, doc_ref_2])

    assert res == [RelationshipReference(nested=False, doc_ref=doc_ref_1),
                   RelationshipReference(nested=False, doc_ref=doc_ref_2)]

    from onto.models.factory import ClsFactory
    ContainsIterable = ClsFactory.create(
        "ContainsIterable",
        schema=ContainsIterableSchema,
        base=FirestoreObject
    )

    obj = ContainsIterable.new(the_iterable=[doc_ref_1, doc_ref_2])
    res = fd.serialize(obj=obj, attr="the_iterable")
    assert res == [RelationshipReference(nested=False, doc_ref=doc_ref_1),
                   RelationshipReference(nested=False, doc_ref=doc_ref_2)]

    obj = ContainsIterable.new(the_iterable={1: doc_ref_1, 2: doc_ref_2})
    res = fd.serialize(obj=obj, attr="the_iterable")
    assert res == {1: RelationshipReference(nested=False, doc_ref=doc_ref_1),
                   2: RelationshipReference(nested=False, doc_ref=doc_ref_2)}


# def test_something():
#
#     class MyDescriptor:
#
#         def __init__(self):
#             self._val = None
#
#         def __get__(self, obj, objtype):
#             return self._val
#
#         def __set__(self, obj, val):
#             self._val = val
#
#     class Owner:
#
#         my = MyDescriptor()
#
#     instance = Owner()
#     instance.my = "hi"
#     assert instance.my == "hi"
#
#     d = dict(a=MyDescriptor())
#     assert d['a'] == "hello"


def test_wrapper():

    class Retriever:

        def __set_name__(self, owner, name):
            self.parent = owner
            self.name = name

        def __set__(self, instance, value):
            raise TypeError("Reassigning an iterable wrapper is not allowed")

        def __get__(self, instance, owner):
            import weakref

            return weakref.proxy(instance.original)

    class Dict(UserDict):
        pass

    class List(UserList):
        pass

    _original = List([1, 2])

    class K:

        proxy = Retriever()

        def __init__(self):
            self.original = _original

    k = K()

    assert k.proxy.index(1) == 0
    assert not hasattr(k.proxy, "non_existent")

    _internal_list = k.proxy
    try:
        k.proxy = list()
    except TypeError as e:
        assert True
    else:
        assert False

    class M:
        pass

    m = M()
    m.original = k.original

    print(m.original)

    assert k.proxy[0] == 1
    assert len(m.original) == 2


def test_proxy(CTX):

    class A(DomainModel):
        foo = attrs.bproperty()

    doc_ref = CTX.db.ref/'A'/'a'
    doc_ref.set(dict(foo='bar'))

    from onto.store import Gallery
    s = Gallery()

    from google.cloud.firestore import transactional
    @transactional
    def execute(transaction):
        rr = RelationshipReference(doc_ref=doc_ref, obj_type=A)
        a = _nest_relationship_import(rr, _store=s)

        s.refresh(transaction=transaction)

        assert a.foo == 'bar'
        assert isinstance(a, A)

    execute(transaction=CTX.db.transaction())

    def execute_no_transaction():
        rr = RelationshipReference(doc_ref=doc_ref, obj_type=A)
        a = _nest_relationship_import(rr, _store=s)

        s.refresh(transaction=None)

        assert a.foo == 'bar'
        assert isinstance(a, A)

    execute_no_transaction()


def test_local_time():
    from onto.mapper.fields import local_time_from_timestamp, timestamp_from_local_time
    local_time_str = local_time_from_timestamp(1545062400)
    assert local_time_str == "2018-12-17T08:00:00"
    local_time_timestamp = timestamp_from_local_time("2018-12-17T08:00:00")
    assert local_time_timestamp == 1545062400


# def test_projected():
#
#     from onto.schema import Schema
#     from onto import fields
#
#     class ArtistSchema(Schema):
#         id = fields.Integer()
#         name = fields.String()
#
#     class AlbumSchema(Schema):
#         id = fields.Projected(ArtistSchema, 'artist')
#
#     in_data = {'id': 42}
#     loaded = AlbumSchema().load(in_data)  # => {'artist': {'id': 42}}
#     print(loaded)
#     dumped = AlbumSchema().dump(loaded)  # => {'id': 42}
#     print(dumped)

#
# def test_init_camelize():
#
#     int_a = fb_fields.Integer(attribute="int_a",
#                     fieldname_mapper=utils.attr_name_to_firestore_key)
#
#     mfunc.assert_called_once_with(
#
#         attribute="int_a",
#
#         dump_to="intA",
#         load_from="intA",
#
#     )




import warnings

from onto.errors import PropertyEvalError
from onto.utils import attr_name_to_firestore_key, \
    firestore_key_to_attr_name
from . import fields
import marshmallow
from marshmallow import post_dump, pre_load, post_load, EXCLUDE


class SchemaMixin:

    f = staticmethod(attr_name_to_firestore_key)
    g = staticmethod(firestore_key_to_attr_name)

    case_conversion = True

    # def on_bind_field(self, field_name, field_obj):
    #     """Hook to modify a field when it is bound to the `Schema`.
    #
    #     No-op by default.
    #     """
    #
    #     if field_obj.attribute is None:
    #         field_obj.attribute = field_name
    #
    #     if field_obj.data_key is None:
    #         if self.case_conversion:
    #             default_data_key = self.f(field_obj.attribute)
    #         else:
    #             default_data_key = field_obj.attribute
    #         field_obj.data_key = default_data_key
    #
    #     # if field_obj.attribute in self.f_mapping:
    #     #     raise ValueError
    #     # else:
    #     #     self.f_mapping[field_obj.attribute] = field_obj.data_key
    #     #
    #     # if field_obj.data_key in self.g_mapping:
    #     #     raise ValueError
    #     # else:
    #     #     self.g_mapping[field_obj.data_key] = field_obj.attribute

    def __init__(self, *args, **kwargs):
        if "unknown" not in kwargs:
            # The assumption made is that if-condition on existence of
            #   key in kwargs is a better design than explicit kwarg = None,
            #   since, as an example, `EXCLUDE is None` may be True, as it
            #   is from an external dependency.
            # TODO: note that unknown passed by positional arg is not caught
            unknown = EXCLUDE
        else:
            unknown = kwargs["unknown"]
        super().__init__(
            *args,
            unknown=unknown,
            **kwargs)
        #
        # self.f_mapping = dict()
        # self.g_mapping = dict()


class BusinessPropertyStoreSchemaMixin():
    pass


class BoilerProperty(object):
    """
    Ref: https://blog.csdn.net/weixin_43265804/article/details/82863984
        content under CC 4.0
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        self.__doc__ = doc

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            try:
                return self.fget(instance)
            except Exception as exception:
                """
                Otherwise AttributeError will be swallowed 
                """
                raise PropertyEvalError from exception

    def __set__(self, instance, value):
        self.fset(instance, value)

    def __delete__(self, instance):
        self.fdel(instance)

    def getter(self, fget):
        return BoilerProperty(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return BoilerProperty(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return BoilerProperty(self.fget, self.fset, fdel, self.__doc__)


class SchemaBase(SchemaMixin, marshmallow.Schema):

    class Meta:
        ordered = True

    @classmethod
    def _get_reserved_fieldnames(cls):
        return set()

    # def get_attribute(self, *args, **kwargs):
    #     # TODO: note this method automatically falls back to super during AttributeError
    #     try:
    #         f = get_attribute_f.get()
    #         return f(self, *args, **kwargs)
    #     except LookupError:
    #         return super().get_attribute(*args, **kwargs)


class BasicSchema(SchemaBase):
    pass


class Schema(SchemaBase):
    """
    Attributes:
    ============
    obj_type: fields.Function
        Exports and imports class name of an instance for differentiating
            different subclasses of PrimaryObject in the same collection.
    TODO: change, now doc_ref
    doc_id: firestore document id
        Caution: probable pitfall; doc_id won't be valid if obj is
            not a subclass of firestore object; watch out for strange
            behaviors such as doc_id being setted twice.
            A possible mistake can be setting doc_id after doc_id is
            read, since two doc_id for the same object can be observed.
    """
    class Meta:
        ordered = True

    obj_type = fields.ObjectTypeField(
        data_key="obj_type",
    )

    _remainder = fields.Remainder(attribute="_remainder", required=False)

    # @pre_load(pass_many=True)
    # def collect_remainder(self, data, many, **kwargs):
    #     _fds = self.fields
    #
    #     field_keys = {field.data_key for key, field in _fds.items()
    #                   if not field.dump_only}
    #
    #     # Keys to set value from keyword arguments
    #     data_keys = data.keys()
    #     remainder_keys = data_keys - field_keys
    #     remainder = {
    #         key: data[key]
    #         for key in remainder_keys
    #     }
    #
    #     if len(remainder) == 0:
    #         return data
    #
    #     for key, val in remainder.items():
    #         del data[key]
    #     data["_remainder"] = remainder
    #
    #     return data
    #
    # @post_dump(pass_many=True)
    # def release_remainder(self, data, many, **kwargs):
    #
    #     if "_remainder" not in data:
    #         return data
    #
    #     if data["_remainder"] is not None:
    #         remainder = data["_remainder"]
    #     else:
    #         remainder = dict()
    #
    #     for key, val in remainder.items():
    #         if key not in data:
    #             data[key] = val
    #     del data["_remainder"]
    #
    #     return data

    @classmethod
    def _get_reserved_fieldnames(cls):
        """ Returns a list of fieldnames to hide when calling
            "to_view_dict" on view model

        :return:
        """
        return {"obj_type", "doc_id", "doc_ref"}


def _get_field(variable_key) -> fields.Field:
    warnings.warn("_get_field is returning wildcard field")
    return fields.Raw(
            load_from=attr_name_to_firestore_key(variable_key)
        )


def _get_field_vars(var_names, fd) -> dict:
    """

    :param var_names: A list of (instance) variable names
    :param fd:
    :return:
    """
    field_vars = dict()
    for var_name in var_names:
        if var_name in fd:
            field_vars[var_name] = _get_field(var_name)
    return field_vars


def _get_instance_variables(obj_cls) -> list:
    """
    Returns a dict of instance variable name: default value
    Note that this method does not work for now
    :param obj_cls:
    :return:
    """
    # init_func: function = obj_cls.__init__
    res = obj_cls.__init__.__code__.co_names
    return list(res)


def generate_schema(obj_cls) -> Schema:
    """
    Generates
    :param obj:
    :return:
    """

    def constructor(self, obj_cls):
        Schema.__init__(self)
        self.obj_cls = obj_cls

    # A list of all fields to serialize and deserialize
    fd = obj_cls.get_fields()

    @post_load
    def make_temp_obj(self, data: dict, **kwargs):
        obj = self.obj_cls()
        for k, v in data.items():
            key, value = firestore_key_to_attr_name(k), v
            if key in fd:
                setattr(obj, key, value)
        return obj

    instance_vars = _get_instance_variables(obj_cls)
    field_vars = _get_field_vars(instance_vars, fd)

    TempSchema = type("TempSchema", (Schema,),
                      {
                          "__init__": constructor,
                          "make_temp_obj": make_temp_obj,
                          **field_vars
                      }
    )

    return TempSchema(obj_cls)



from onto.context import Context as CTX
# from google.cloud.firestore import CollectionReference

from onto.database import Reference
from onto.firestore_object import FirestoreObject
from onto.models.meta import SerializableMeta


class CollectionMemberMeta(SerializableMeta):

    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)
        if hasattr(klass, "Meta"):
            meta = klass.Meta
            if hasattr(meta, "collection_name"):
                klass._collection_name = meta.collection_name
        return klass


try:
    from pony.orm.core import EntityMeta
    class with_pony(EntityMeta, CollectionMemberMeta):

        def __new__(mcs, name, bases, attrs):
            db = CTX.dbs.pony
            bases = (*bases, db.Entity,)
            klass = super().__new__(mcs, name, bases, attrs)
            return klass

        def __init__(mcs, name, bases, attrs):
            db = CTX.dbs.pony
            bases = (*bases, db.Entity,)
            super().__init__(name, bases, attrs)
except ImportError:
    import warnings
    warnings.warn("with_pony skipped")


class CollectionMixin:

    _collection_name = None

    @property
    def collection_name(self):
        """ Returns the root collection name of the class of objects.
                If cls._collection_name is not specified, then the collection
                name will be inferred from the class name. Note that different
                types of objects may share one collection.
        """
        # if type(self) == FirestoreObject:
        #     raise ValueError("collection_name is read from class name, "
        #                      "only subclass is supported. ")
        return self._get_collection_name()

    @classmethod
    def _get_collection_name(cls):
        if cls._collection_name is None:
            cls._collection_name = cls.__name__
        return cls._collection_name

    @property
    def collection(self):
        """ Returns the firestore collection of the current object
        """
        return self._get_collection()

    @classmethod
    def _get_collection(cls) -> Reference:
        return cls._datastore().ref / cls._get_collection_name()

    @classmethod
    def ref_from_id(cls, doc_id):
        """ Returns a Document Reference from doc_id supplied.

        :param doc_id: Document ID
        """
        return cls._get_collection() / doc_id



"""
PORTED

Note: C-implemented functions don't show up as Python stack frames and are thus hidden to this script. This will
result in false positives. Consider Python function f() which calls a = g(). g() is C-implemented and calls b = f2().
When f2() tries to lookup up the assigned name, it will get a instead of b because the script is oblivious to C
functions. (At least this is how I guess it will work :P )

Ref: https://stackoverflow.com/questions/8875202/can-an-object-inspect-the-name-of-the-variable-its-been-assigned-to
"""
import inspect
import dis


def take1(iterator):
    try:
        return next(iterator)
    except StopIteration:
        raise Exception("missing bytecode instruction") from None


def take(iterator, count):
    for x in range(count):
        yield take1(iterator)


def get_assigned_name(frame):
    """Takes a frame and returns a description of the name(s) to which the
    currently executing CALL_FUNCTION instruction's value will be assigned.

    fn()                    => None
    a = fn()                => "a"
    a, b = fn()             => ("a", "b")
    a.a2.a3, b, c* = fn()   => ("a.a2.a3", "b", Ellipsis)
    """

    iterator = iter(dis.get_instructions(frame.f_code))
    for instr in iterator:
        if instr.offset == frame.f_lasti:
            break
    else:
        assert False, "bytecode instruction missing"

    if instr.opname.startswith('CALL_'):
        instr = take1(iterator)
        if instr.opname == 'POP_TOP':
            raise ValueError("not assigned to variable")
        return instr_dispatch(instr, iterator)
    else:
        return instr_dispatch(instr, iterator)


def instr_dispatch(instr, iterator):
    opname = instr.opname
    if (opname == 'STORE_FAST'              # (co_varnames)
            or opname == 'STORE_GLOBAL'     # (co_names)
            or opname == 'STORE_NAME'       # (co_names)
            or opname == 'STORE_DEREF'):    # (co_cellvars++co_freevars)
        return instr.argval
    if opname == 'UNPACK_SEQUENCE':
        return tuple(instr_dispatch(instr, iterator)
                     for instr in take(iterator, instr.arg))
    if opname == 'UNPACK_EX':
        return (*tuple(instr_dispatch(instr, iterator)
                     for instr in take(iterator, instr.arg)),
                Ellipsis)
    # Note: 'STORE_SUBSCR' and 'STORE_ATTR' should not be possible here.
    # `lhs = rhs` in Python will evaluate `lhs` after `rhs`.
    # Thus `x.attr = rhs` will first evalute `rhs` then load `a` and finally
    # `STORE_ATTR` with `attr` as instruction argument. `a` can be any
    # complex expression, so full support for understanding what a
    # `STORE_ATTR` will target requires decoding the full range of expression-
    # related bytecode instructions. Even figuring out which `STORE_ATTR`
    # will use our return value requires non-trivial understanding of all
    # expression-related bytecode instructions.
    # Thus we limit ourselfs to loading a simply variable (of any kind)
    # and a arbitary number of LOAD_ATTR calls before the final STORE_ATTR.
    # We will represents simply a string like `my_var.loaded.loaded.assigned`
    if opname in {'LOAD_CONST', 'LOAD_DEREF', 'LOAD_FAST',
                    'LOAD_GLOBAL', 'LOAD_NAME'}:
        return instr.argval + "." + ".".join(
            instr_dispatch_for_load(instr, iterator))
    raise NotImplementedError("assignment could not be parsed: "
                              "instruction {} not understood"
                              .format(instr))


def instr_dispatch_for_load(instr, iterator):
    instr = take1(iterator)
    opname = instr.opname
    if opname == 'LOAD_ATTR':
        yield instr.argval
        yield from instr_dispatch_for_load(instr, iterator)
    elif opname == 'STORE_ATTR':
        yield instr.argval
    else:
        raise NotImplementedError("assignment could not be parsed: "
                                  "instruction {} not understood"
                                  .format(instr))


from typing import List, Optional, Dict

import typing

from onto import attrs
from onto.attrs.attribute_new import AttributeMixin, AttributeBase
from onto.domain_model import DomainModel


from onto.attrs import unit


def test_something():

    initial = unit.Monad(decor=unit.DecoratorBase.get_root())
    res = initial.of_type(type_cls=str).import_required

    fd = res.decor._marshmallow_field_constructor()
    from marshmallow import fields
    assert str(fd) == str(fields.String(required=True))

    def fget(self):
        return 'hello'

    _1 = res.getter(fget)

    assert _1.decor.fget == fget
    assert _1.decor.fget(self=None) == 'hello'


def test_in_class():

    initial = unit.Monad(decor=unit.DecoratorBase.get_root())

    class A:
        status = initial.of_type(type_cls=str)

        @status.getter
        def status(self):
            return 'ok'

    assert A.status.decor.fget(A()) == 'ok'


def test_list():

    initial = unit.Monad(decor=unit.DecoratorBase.get_root())

    class A:
        status_list = initial.list(
            value=initial.of_type(type_cls=str).import_required
        )

        @status_list.getter
        def status_list(self):
            return ['hello', 'world']

    assert A.status_list.decor.fget(A()) == ['hello', 'world']


def test_decorator():

    # _1 = unit.OfType(type_cls=str, decorated=None)
    # _2 = unit.ImportRequired(decorated=_1)
    #
    # from marshmallow import fields
    #
    # assert _2._marshmallow_field_cls == fields.String
    # assert _2._marshmallow_required
    # fd =_2._marshmallow_field_constructor()
    # assert str(fd) == str(fields.String(required=True))
    #
    # def fget(self):
    #     return 'hello'
    #
    # _3 = unit.Getter(fget=fget, decorated=_2)
    # assert _3.fget == fget

    attrs = AttributeBase.get_root()

    status = attrs.str.optional.data_key('status')

    # num_meetings = attrs.int.optional
    # meetings = attrs.list(value=attrs.str)

    from onto.mapper import fields
    from marshmallow.fields import missing_

    # assert str(fd) == str(fields.String(data_key='my_status', required=False))
    kwargs = status.decor._marshmallow_field_kwargs
    assert dict(kwargs) == {
        'data_key': 'status'
    }

    fd2 = status.marshmallow_field
    assert str(fd2) == str(fields.String(data_key='status', required=False, missing=str))

    monad3_1 = unit.MonadContext.context().data_key('my_status')
    with monad3_1:
        monad3_2_1 = unit.MonadContext.context().optional
        with monad3_2_1:
            fd3 = status._marshmallow_field_constructor()
            assert str(fd3) == str(fields.String(data_key='my_status', required=False, missing=str))
        monad3_2_2 = unit.MonadContext.context().data_key('the_status')
        with monad3_2_2:
            fd3 = status._marshmallow_field_constructor()
            assert str(fd3) == str(fields.String(data_key='the_status', required=False, missing=str))


def test_domain_model():

    attrs = AttributeBase.get_root()

    class Meeting(DomainModel):
        class Meta:
            collection_name = "meetings"
            case_conversion = False

        @classmethod
        def _datastore(cls):
            from onto.database import Database
            return Database()

        # location = attrs.relation.not_nested.dm_cls('Location')
            # (nested=False, dm_cls='Location')
        # users: Dict['Ticket', 'User'] = attrs.not_nested
        # users = users.dict(
        #     key=attrs.of_type('Ticket'),
        #     value=attrs.of_type('User').optional
        # )
        #     # (nested=False, dm_cls='User',
        #     #                    collection=list)
        # tickets: List['Ticket'] = attrs.relation.not_nested
        status: str = attrs

        num_meetings: int = attrs.int.optional.default_value(-1)  # TODO: remove default_value
        @num_meetings.getter
        def num_meetings(self):
            return len(self.meetings)

        meetings: typing.List[typing.AnyStr] = attrs.list(
            value=lambda at: at.str
        )

    # base1 = unit.DataKey(data_key='my_status', decorated=None)
    #
    # with decor_base(decor=base1):
    #     base2 = unit.ImportEnabled(base1)
    #     with decor_base(decor=base2):
    #         fd = Meeting.status._marshmallow_field_constructor()
    #
    from onto.mapper import fields

    # assert str(fd) == str(fields.String(data_key='my_status', required=False))
    context = unit.MonadContext.context().marshmallow_capable_base().string
    with context:
        fd2 = Meeting.status.marshmallow_field
    assert str(fd2) == str(fields.String(data_key='status', required=False))

    monad3_1 =context.data_key('my_status')
    with monad3_1:
        monad3_2_1 = unit.MonadContext.context().optional
        with monad3_2_1:
            fd3 = Meeting.status._marshmallow_field_constructor()
            assert str(fd3) == str(fields.String(data_key='my_status', required=False))
        monad3_2_2 = context.data_key('the_status')
        with monad3_2_2:
            fd3 = Meeting.status._marshmallow_field_constructor()
            assert str(fd3) == str(fields.String(data_key='the_status', required=False))

    # with decor_base(decor=base1):
    #     base2 = unit.ImportEnabled(base1)
    #     with decor_base(decor=base2):
    #         fd = Meeting.status._marshmallow_field_constructor()

    m = Meeting.new(status='ok', meetings=['meeting one', 'meeting_two'])
    assert m.status == 'ok'
    from testfixtures import compare, MappingComparison
    compare(m.to_dict(), MappingComparison({
        'status': 'ok',
        'num_meetings': 2,
        'meetings': ['meeting one', 'meeting_two']
    }, partial=True))

    from onto.models.utils import _graphql_object_type_from_attributed_class
    res = _graphql_object_type_from_attributed_class(Meeting, input=False)
    assert res


def test_graphql_custom_scalar_type():

    from onto.models.base import Serializable
    class MyString(Serializable):
        class Meta:
            unwrap = 'value'

        value: str = attrs.attrs.nothing

    class A1(DomainModel):
        message = attrs.attrs.embed(MyString)

    from onto.models.utils import _graphql_object_type_from_attributed_class
    res = _graphql_object_type_from_attributed_class(A1, input=False)
    assert str(res) == 'A1'
    assert A1.from_dict({'value': 'hello'}).message != 'hello'


# def test_monad():



#
# class Meeting(DomainModel):
#
#     class Meta:
#         collection_name = "meetings"
#
#     location = attrs.relation.not_nested.dm_cls('Location')
#         # (nested=False, dm_cls='Location')
#     users: Dict['Ticket', 'User'] = attrs.not_nested
#     users = users.dict(
#         key=attrs.of_type('Ticket'),
#         value=attrs.of_type('User').optional
#     )
#         # (nested=False, dm_cls='User',
#         #                    collection=list)
#     tickets: List['Ticket'] = attrs.relation.not_nested
#     status = attrs.string
#
#     with attrs.read_only:
#         num_meetings = attrs.integer.optional
#
#
# class Ticket(DomainModel):
#     role = attrs.string.required
#     user: Optional['User'] = attrs.relation.not_nested
#     attendance = attrs.bproperty(type_cls=bool)
#
#     meetings = attrs.relation(import_required=False, dm_cls='Meeting', collection=list, nested=False)



from .models import City
from onto.context import Context as CTX
from onto.view_model import ViewModel, Struct
from onto.store import Store, reference
from onto import attrs


class CityStore(Store):
    city = reference(many=False)


class CityView(ViewModel):

    name = attrs.bproperty()
    country = attrs.bproperty()

    @classmethod
    def new(cls, snapshot):
        store = CityStore()
        store.add_snapshot("city", dm_cls=City, snapshot=snapshot)
        store.refresh()
        return cls(store=store)

    @name.getter
    def name(self):
        return self.store.city.city_name

    @country.getter
    def country(self):
        return self.store.city.country

    @property
    def doc_ref(self):
        return CTX.db.document(f"cityView/{self.store.city.doc_id}")



"""
DELICATE: implementation depends on magic; make changes carefully
Interface was inspired by public API of RethinkDB.
RethinkDB is released under apache 2.0 license.
"""


class CMP:

    def __getattr__(self, item):
        cond = RootCondition()
        from onto.context import Context as CTX
        cond.comparator_cls = CTX.db.Comparators
        cond.attr_name = item
        return cond


class Condition:
    """
    Maps <, <=, ==, >=, >, and in to python comparator.
    See test_query for usage
    """

    @property
    def comparator_cls(self):
        return self._comparator_cls

    @comparator_cls.setter
    def comparator_cls(self, value):
        self._comparator_cls = value

    @property
    def attr_name(self):
        return self._attr_name

    @attr_name.setter
    def attr_name(self, value):
        self._attr_name = value

    def _constraint_for(self, other, opname):
        op = getattr(self.comparator_cls, opname)
        return [*self.constraints, (op, other)]

    def __and__(self, other):
        return super().__and__(self, other)

    def __iand__(self, other):
        return super().__iand__(self, other)

    # def __bool__(self):
    #     """
    #     So that __and__ is triggered;
    #     Note that this affects __or__
    #     :return:
    #     """
    #     return False


class RootCondition(Condition):

    def __init__(self, *args, attr_name=None, comparator_cls=None, **kwargs):
        self.attr_name = attr_name
        self.comparator_cls = comparator_cls
        super().__init__(*args, **kwargs)

    def __lt__(self, other):
        return self._descendant(constraints=self._constraint_for(other, 'lt'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __le__(self, other):
        return self._descendant(constraints=self._constraint_for(other, 'le'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __gt__(self, other):
        return self._descendant(constraints=self._constraint_for(other, 'gt'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __ge__(self, other):
        return self._descendant(constraints=self._constraint_for(other, 'ge'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __eq__(self, other):
        return self._descendant(constraints=self._constraint_for(other, 'eq'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    # def __bool__(self):
    #     """
    #     So that __and__ is triggered;
    #     Note that this affects __or__
    #     :return:
    #     """
    #     return True
    #
    # def __and__(self, other):
    #     return self._descendant(constraints=[*self.constraints, *other.constraints], comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __ne__(self, other):
        raise ValueError("Not equal is not supported for Firestore")

    def has(self, item):
        """
        Maps to array membership "in"
        (needs to be reversed)

        :param item:
        :return:
        """
        return self._descendant(constraints=self._constraint_for(item, 'contains'), comparator_cls=self.comparator_cls, attr_name=self.attr_name)


    @property
    def _descendant(self):
        return NodeCondition

    @property
    def constraints(self):
        return []

    @Condition.comparator_cls.getter
    def comparator_cls(self):
        if self._comparator_cls is None:
            return self.parent._datastore().Comparators
        else:
            return self._comparator_cls

    @Condition.attr_name.getter
    def attr_name(self):
        if self._attr_name is None:
            return self.name
        else:
            return self._attr_name


class NodeCondition(Condition):

    @property
    def _descendant(self):
        return NodeCondition

    def __init__(self, *args, constraints=None, comparator_cls=None, attr_name=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.constraints = constraints
        self.comparator_cls = comparator_cls  # This calls setter in superclass
        self.attr_name = attr_name  # This calls setter in superclass

    def __lt__(self, other):
        self.constraints = self._constraint_for(other, 'lt')
        return self

    def __le__(self, other):
        self.constraints = self._constraint_for(other, 'le')
        return self

    def __gt__(self, other):
        self.constraints = self._constraint_for(other, 'gt')
        return self

    def __ge__(self, other):
        self.constraints = self._constraint_for(other, 'ge')
        return self

    def __eq__(self, other):
        self.constraints = self._constraint_for(other, 'eq')
        return self

    # #
    # def __and__(self, other):
    #     self._descendant(constraints=[*self.constraints, *other.constraints], comparator_cls=self.comparator_cls, attr_name=self.attr_name)

    def __ne__(self, other):
        raise ValueError("Not equal is not supported for Firestore")

    def has(self, item):
        """
        Maps to array membership "in"
        (needs to be reversed)

        :param item:
        :return:
        """
        self.constraints = self._constraint_for(item, 'contains')
        return self


v = CMP()



from .firestore import FirestoreSink as firestore, \
    ViewModelSink as view_model, FormSink as form, \
    DomainModelSink as domain_model
from .rest import rest_client, rest_server
from .ws import websocket




from onto.context import Context as CTX
from .base import Sink


class LcSink(Sink):

    def emit(self, ref, snapshot):
        CTX.dbs.leancloud.set(ref=ref, snapshot=snapshot)


class ViewModelSink(Sink):

    def emit(self, obj):
        CTX.dbs.leancloud.set(ref=obj.doc_ref, snapshot=obj.to_snapshot())



 from onto.primary_object import PrimaryObject


class Entity(PrimaryObject):
    """
    Entity is for flink stateful functions
    """

    @classmethod
    def _datastore(cls):
        from onto.context import Context as CTX
        return CTX.db


import statefun

functions = statefun.StatefulFunctions()

handler = statefun.RequestReplyHandler(functions)


from dataclasses import dataclass


def register_root_entity(entity):

    functions.register()


def method_invoked(context, call_event: CallEvent):
    state = context.state('_internal_state').unpack(InternalState)
    if not state:
        state = SeenCount()
        state.seen = 1
    else:
        state.seen += 1
    context.state('seen_count').pack(state)

    response = compute_greeting(greet_request.name, state.seen)

    egress_message = kafka_egress_record(topic="greetings", key=greet_request.name, value=response)
    context.pack_and_send_egress("example/greets", egress_message)


class MethodSerializer:
    """
    (Internal)

    Ref: https://stackoverflow.com/questions/5103735/better-way-to-log-method-calls-in-python
    """

    def _decorator(self, f):
        import functools
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            handler()
            context.send(typename=self.doc_ref.collection, id=self.doc_ref.id, message=Any)
            return (f.__name__, args, kwargs)
        return wrapper

    def __getattribute__(self, item):

        # def f(*args, **kwargs):
        #     raise ValueError('Should not have been called')

        value = object.__getattribute__(self, item)
        if callable(value):
            decorator = object.__getattribute__(self, '_decorator')
            return decorator(value)
        return value


class EntityProxy(PrimaryObject):
    """


    Useful: https://docs.microsoft.com/en-us/dotnet/architecture/microservices/architect-microservice-container-applications/communication-in-microservice-architecture
    """

    pass

        #
        # statefun.kafka_egress_record()
        # func_addr






try:
    """ DocumentMediator listens to change in a single document and triggers
            functions when such document change.
    """
    from .document import ViewMediatorDAV as DocumentMediator

    """ QueryMediator listens to the result of a query and triggers 
            functions when the result of such query changes. 
    """
    from .query_delta import ViewMediatorDeltaDAV as QueryMediator, \
        ProtocolBase, \
        OnSnapshotTasksMixin
except ImportError:
    # TODO: make better
    import warnings
    warnings.warn("query_delta not imported")

try:

    """ RestMediator pulls and modifies data when requested with REST API. 
    """
    from .rest_api import ViewMediator as RestMediator
except ImportError:
    # TODO: make better
    import warnings
    warnings.warn("rest_api not imported")

try:
    """ WsMediator establishes websocket connection with client. 
    """
    from .websocket import ViewMediatorWebsocket as WsMediator
except ImportError:
    # TODO: make better
    import warnings
    warnings.warn("websocket not imported")

class Mediator:

    source = None
    subscribe_user_view = None

    @classmethod
    def start(cls):
        cls.source.start()



from onto.mapper.fields import OBJ_TYPE_ATTR_NAME
from onto.utils import snapshot_to_obj
from onto.context import Context as CTX

# from google.cloud import firestore

def is_fb_snapshot(snapshot: 'google.cloud.firestore.DocumentSnapshot') -> bool:
    """
    Returns true if a snapshot is generated by flask-boiler
        (checks if obj_type exists)

    :param snapshot:
    :return:
    """
    try:
        _ = snapshot.get("obj_type")
        return True
    except KeyError as ke:
        return False


def convert_query_ref(func):
    """
    Converts a generator of firestore DocumentSnapshot's to a generator
            of objects.

    :param super_cls:
    """
    def call(cls, *args, **kwargs):
        query_ref = func(cls, *args, **kwargs)
        for res in query_ref.stream():
            assert isinstance(res, firestore.DocumentSnapshot)
            # if is_fb_snapshot(snapshot=res):
            #     # Skip this snapshot if it is not generated by flask-boiler
            from onto.database.firestore import FirestoreSnapshot
            yield snapshot_to_obj(snapshot=res, super_cls=cls)
    return call


def convert_query(func):
    def call(cls, *args, **kwargs):
        q, db = func(cls, *args, **kwargs)
        for ref, snapshot in db.query(q):
            yield snapshot_to_obj(
                reference=ref,
                snapshot=snapshot,
                super_cls=cls)
    return call


class QueryMixin:

    @classmethod
    @convert_query
    def all(cls):
        """ Gets object that is a subclass of the current cls in
                the collection.
        """

        return cls.get_query(), cls._datastore()

    @classmethod
    @convert_query
    def where(cls, *args, **kwargs):
        return cls.get_query().where(*args, **kwargs), cls._datastore()

    @classmethod
    def get_obj_type_condition(cls):
        schema_obj = cls.get_schema_obj()
        if OBJ_TYPE_ATTR_NAME not in schema_obj.fields:
            return None
        else:
            return schema_obj.fields[OBJ_TYPE_ATTR_NAME]\
                .get_obj_type_condition(cls)

    @classmethod
    def get_query(cls):
        """ Returns a query with parent=cls._get_collection(), and
                limits to obj_type of subclass of cls.
        """

        from onto.query.query import DomainModelQuery
        return DomainModelQuery(parent=cls)

    # @classmethod
    # @convert_query_ref
    # def where(cls, *args,
    #           acsending=None,
    #           descending=None,
    #           end_at=None,
    #           end_before=None,
    #           limit=None,
    #           offset=None,
    #           order_by=None,
    #           select=None,
    #           start_after=None,
    #           start_at=None,
    #           **kwargs,):
    #     """ Queries the datastore. Note that indexes may need to be added
    #             from the link provided by firestore in the error messages.
    #
    #     NOTE: all subclasses of the current class will be queried
    #     NOTE: this method will fail when the descendents of the class
    #         plus itself counts to >10 due to Firestore query limitation.
    #         (10 "in" queries max, which is 9 descendant classes max).
    #         Watch for dynamically constructed class
    #         in runtime, as this may increase the number of descendants of its
    #         superclass, and may result in fail. To turn off obj_type behavior,
    #         use a separate collection for each class and/or customize
    #         obj_type field/attribute.
    #
    #     TODO: post more examples for customizing obj_type field/attribute
    #
    #     TODO: add error handling and argument checking
    #     TODO: implement limit, orderby, etc.
    #
    #     :param args:
    #     :param kwargs:
    #     """
    #
    #     cur_where = cls._where_query(*args, **kwargs,
    #                                  cur_where=cls.get_query())
    #
    #     if limit is not None:
    #         cur_where = cur_where.limit(count=limit)
    #
    #     return cur_where




from collections import namedtuple
from functools import cached_property

from flasgger import SwaggerView
from flask import request, jsonify

from onto.attrs import attrs
from onto.view.base import ViewMediatorBase

from onto.models.base import Serializable


class ParamsParams(Serializable):
    page_size = attrs.int
    current = attrs.int

    def append_conditions(self, q):
        return q

class Sort(Serializable):
    foo = attrs.string
    def append_conditions(self, q):
        return q


class Filter(Serializable):
    foo = attrs.string
    def append_conditions(self, q):
        return q


class Params(Serializable):
    params = attrs.embed(ParamsParams)
    sort = attrs.embed(Sort)
    filter = attrs.embed(Filter)


PaginatedResponse = namedtuple('PaginatedResponse', ['success', 'data', 'total'], defaults=[True, None, None])


class ViewMediator(ViewMediatorBase):
    """
    Registers a REST API Resource for a view model
    """

    def __init__(self,
                 view_model_cls=None,
                 mutation_cls=None,
                 *args,
                 app=None,
                 **kwargs):
        """ Initialize a ViewMediator for REST API

        :param view_model_cls
        :param mutation_cls: a subclass of Mutation to handle the changes
                POST, PATCH, UPDATE, PUT, DELETE made to the list of view
                models or a single view model.
        :param args:
        :param app: Flask App
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.view_model_cls = view_model_cls
        self.mutation_cls = mutation_cls
        self.app = app
        self.rule_view_cls_mapping = dict()
        self.default_tag = self.view_model_cls.__name__

    @cached_property
    def paginated_query_cls(_self):
        from onto.models.base import Serializable
        from onto.attrs import attrs
        return type(
            f'{_self.view_model_cls.__name__}PaginatedQueryResponse',
            (Serializable,),
            {
                'total': attrs.integer,
                'data': attrs.list(value=attrs.embed(_self.view_model_cls)),
                'success': attrs.bool
            }
        )

    def add_instance_get(self, rule=None, instance_get_view=None):
        """
        Add GET operation for resource/instance_id

        :param rule: flask url rule
        :param instance_get_view: Flask View
        :return:
        """

        if instance_get_view is None:
            instance_get_view = self._default_instance_get_view()

        name = self.view_model_cls.__name__ + "GetView"
        assert rule is not None
        self.app.add_url_rule(
        rule,
        view_func=instance_get_view.as_view(name=name),
        methods=['GET']
        )

        self.rule_view_cls_mapping[(rule, 'GET')] = instance_get_view

    def add_list_get(self, rule=None, list_get_view=None):
        """
        Add GET operation for resource to get a list of instances

        :param rule: flask url rule
        :param list_get_view: Flask View
        :return:
        """
        if list_get_view is None:
            list_get_view = self._default_list_get_view()

        name = self.view_model_cls.__name__ + "ListGetView"
        assert rule is not None
        self.app.add_url_rule(
        rule,
        view_func=list_get_view.as_view(name=name),
        methods=['GET']
        )
        self.rule_view_cls_mapping[(rule, 'GET')] = list_get_view

    def add_list_post(self, rule=None, list_post_view=None):
        """
        Add POST operation for resource to add an instance to a list
                of instances

        :param rule: flask url rule
        :param list_post_view: Flask View
        :return:
        """
        if list_post_view is None:
            list_post_view = self._default_list_post_view()
        name = self.view_model_cls.__name__ + "ListPostView"
        assert rule is not None
        self.app.add_url_rule(
            rule,
            view_func=list_post_view.as_view(name=name),
            methods=['POST']
        )
        self.rule_view_cls_mapping[(rule, 'POST')] = list_post_view

    def add_list_post_many(self, rule=None, list_post_view=None):
        """
        Add POST operation for resource to many instances to a list
                of instances

        :param rule: flask url rule
        :param list_post_view: Flask View
        :return:
        """
        if list_post_view is None:
            list_post_view = self._default_list_post_many_view()
        name = self.view_model_cls.__name__ + "ListPostManyView"
        assert rule is not None
        self.app.add_url_rule(
            rule,
            view_func=list_post_view.as_view(name=name),
            methods=['POST']
        )
        self.rule_view_cls_mapping[(rule, 'POST')] = list_post_view

    def add_instance_put(self, rule=None, instance_put_view=None):
        """
        Add POST operation for resource to add an instance to a list
                of instances

        :param rule: flask url rule
        :param list_post_view: Flask View
        :return:
        """
        if instance_put_view is None:
            instance_put_view = self._default_instance_put_view()
        name = self.view_model_cls.__name__ + "InstancePutView"
        assert rule is not None
        self.app.add_url_rule(
            rule,
            view_func=instance_put_view.as_view(name=name),
            methods=['PUT']
        )
        self.rule_view_cls_mapping[(rule, 'PUT')] = instance_put_view

    def add_instance_patch(self, rule=None, instance_patch_view=None):
        """
        Add PATCH operation for making changes to an instance

        :param rule: flask url rule
        :param instance_patch_view: Flask View
        :return:
        """

        if instance_patch_view is None:
            instance_patch_view = self._default_instance_patch_view()

        name = self.view_model_cls.__name__ + "InstancePatchView"
        assert rule is not None
        self.app.add_url_rule(
            rule,
            view_func=instance_patch_view.as_view(name=name),
            methods=['PATCH']
        )
        self.rule_view_cls_mapping[(rule, 'PATCH')] = instance_patch_view

    def add_instance_delete(self, rule=None, instance_delete_view=None):
        """ Add DELETE operation for making changes to an instance

        :param rule:
        :param instance_patch_view: Flask View
        :return:
        """

        if instance_delete_view is None:
            instance_delete_view = self._default_instance_delete_view()

        name = self.view_model_cls.__name__ + "InstanceDeleteView"
        assert rule is not None
        self.app.add_url_rule(
            rule,
            view_func=instance_delete_view.as_view(name=name),
            methods=['DELETE']
        )
        self.rule_view_cls_mapping[(rule, 'DELETE')] = instance_delete_view

    def _default_list_post_view(_self):
        """ Returns a default flask view to handle POST to a list of instances

        """

        class PostView(SwaggerView):

            tags = [_self.default_tag]

            description = "Adds an instance to a list "

            responses = {
                    200: {
                        "description": description,
                        "schema": _self.view_model_cls.get_schema_cls()
                    }
                }

            parameters = [
                    {
                        "name": "doc_id",
                        "in": "path",
                        "type": "string",
                        "required": True,
                    },
                {
                    "name": "body",
                    "in": "body",
                    "required": True,
                    "schema": _self.view_model_cls.get_schema_cls(),
                }
                ]

            def post(self, *args, **kwargs):
                obj = _self.mutation_cls.mutate_create(
                    **kwargs,
                    data=request.json
                )
                if obj is not None:
                    return jsonify(obj.to_dict())
                else:
                    from flask import Response
                    return Response(status=204)
        return PostView

    def _default_list_post_many_view(_self):
        """ Returns a default flask view to handle POST to a list of instances

        """

        class PostManyView(SwaggerView):

            tags = [_self.default_tag]

            description = "Adds many instances to a list "

            responses = {
                    204: {
                        "description": description,
                        "schema": None
                    }
                }

            parameters = [
                    {
                        "name": "doc_id",
                        "in": "path",
                        "type": "string",
                        "required": True,
                    },
                {
                    "name": "body",
                    "in": "body",
                    "required": True,
                    "schema": {
                        'type': 'array',
                        'items': _self.view_model_cls.get_schema_cls(),
                    }
                }
                ]

            def post(self, *args, **kwargs):
                obj = _self.mutation_cls.mutate_create_many(
                    **kwargs,
                    data=request.json
                )
                if obj is not None:
                    return jsonify(obj.to_dict())
                else:
                    from flask import Response
                    return Response(status=204)
        return PostManyView

    def _default_instance_patch_view(_self):
        """ Returns a default flask view to handle PATCH to an instance

        """
        # TODO: change to dynamically construct class to avoid class
        #           name conflict

        class PatchView(SwaggerView):

            tags = [_self.default_tag]

            description = "Patches a specific instance "

            responses = {
                200: {
                    "description": description,
                    "schema": _self.view_model_cls.get_schema_cls()
                }
            }

            parameters = [
                {
                    "name": "doc_id",
                    "in": "path",
                    "type": "string",
                    "required": True,
                },
                {
                    "name": "body",
                    "in": "body",
                    "required": True,
                    "schema": _self.view_model_cls.get_schema_cls(),
                }
            ]

            def patch(self, doc_id=None, **kwargs):
                d = {
                    _self.view_model_cls.get_schema_cls().g(key): val
                    for key, val in request.json.items()
                }
                _self.mutation_cls.mutate_patch(
                    doc_id=doc_id, data=d)
                # time.sleep(1)  # TODO: delete after implementing sync
                return jsonify({
                    "operation_status": "success"
                })

        return PatchView

    def _default_instance_put_view(_self):
        """ Returns a default flask view to handle PUT to an instance

        """

        # TODO: change to dynamically construct class to avoid class
        #           name conflict

        class PutView(SwaggerView):
            tags = [_self.default_tag]

            description = "PUT a specific instance "

            responses = {
                200: {
                    "description": description,
                    "schema": _self.view_model_cls.get_schema_cls()
                }
            }

            parameters = [
                {
                    "name": "doc_id",
                    "in": "path",
                    "type": "string",
                    "required": True,
                },
                {
                    "name": "body",
                    "in": "body",
                    "required": True,
                    "schema": _self.view_model_cls.get_schema_cls(),
                }
            ]

            def put(self, doc_id=None, **kwargs):
                obj = _self.mutation_cls.mutate_put(
                    doc_id=doc_id,
                    data=request.json,
                    **kwargs,
                )
                if obj is not None:
                    return jsonify(obj.to_dict())
                else:
                    from flask import Response
                    return Response(status=200)
        return PutView

    def _default_instance_delete_view(_self):
        """ Returns a default flask view to handle PATCH to an instance

        """
        # TODO: change to dynamically construct class to avoid class
        #           name conflict

        class DeleteView(SwaggerView):

            tags = [_self.default_tag]

            description = "Deletes a specific instance "

            responses = {
                200: {
                    "description": description,
                    "schema": _self.view_model_cls.get_schema_cls()
                }
            }

            parameters = [
                {
                    "name": "doc_id",
                    "in": "path",
                    "type": "string",
                    "required": True,
                }
            ]

            def delete(self, doc_id=None, **kwargs):
                _self.mutation_cls.mutate_delete(
                    doc_id=doc_id)
                # time.sleep(1)  # TODO: delete after implementing sync
                return jsonify({
                    "operation_status": "success"
                })

        return DeleteView

    def _default_instance_get_view(_self):
        """ Returns a default flask view to handle GET to an instance

        """
        # TODO: change to dynamically construct class to avoid class
        #           name conflict

        class GetView(SwaggerView):

            tags = [_self.default_tag]

            description = "Gets an instance "

            responses = {
                200: {
                    "description": description,
                    "schema": _self.view_model_cls.get_schema_cls()
                }
            }

            parameters = [
                {
                    "name": "doc_id",
                    "in": "path",
                    "type": "string",
                    "required": True,
                }
            ]

            def get(self, *args, **kwargs):
                instance = _self.view_model_cls.get(*args, **kwargs)
                # time.sleep(1)  # TODO: delete after implementing sync
                if isinstance(instance, dict):
                    return jsonify(instance)
                else:
                    return jsonify(instance.to_view_dict())

        return GetView

    def _default_list_get_view(_self):

        class ListGetView(SwaggerView):
            tags = [_self.default_tag]

            description = "Query instance by conditions"

            responses = {
                200: {
                    "description": description,
                    "schema": _self.paginated_query_cls.get_schema_cls()
                }
            }

            parameters = [
                {
                    "name": "params",
                    "in": "path",
                    "type": "object",
                    "required": False,
                },
                {
                    "name": "sort",
                    "in": "path",
                    "type": "object",
                    "required": False,
                },
                {
                    "name": "filter",
                    "in": "path",
                    "type": "object",
                    "required": False,
                },
            ]

            def get(self):
                from flask import request
                params_str = request.args.get('params', '{}')
                sort_str = request.args.get('sorter', '{}')
                filter_str = request.args.get('filter', '{}')
                import json

                params_d = json.loads(params_str)
                sort_d = json.loads(sort_str)
                filter_d = json.loads(filter_str)
                params_cls = getattr( _self.view_model_cls, 'params', Params )
                params = params_cls.from_dict({
                    'params': params_d,
                    'sort': sort_d,
                    'filter': filter_d,
                })

                res = _self.view_model_cls.get_many(params=params)
                from collections import namedtuple
                if isinstance(res, PaginatedResponse):
                    data = res.data
                    total = res.total
                    success = res.success
                else:
                    data = list(res)
                    total = 0
                    success = True
                paginated = _self.paginated_query_cls.new(
                    data=data,
                    total=total,
                    success=success
                )
                from flask import jsonify
                return jsonify(paginated.to_dict())

        return ListGetView




"""
Reference: flask-restful docs
"""

import warnings
from functools import wraps, partial

from flask import request, Response
from flask_restful import abort

from .context import Context

# Flag for whether main app is in TESTING mode
def is_testing():
    testing = Context.config.TESTING
    if testing is True:
        warnings.warn("Testing mode is True. Set to False before release. ")
    return testing


# For Testing Purposes. If not in TESTING mode, the function verfifies the token with FirebaseApp.
# Otherwise, a fixed uid will be returned in the format {"uid":"testuid1"}

class AuthVerifyIdTokenFunc:

    def __call__(self, *args, **kwargs):
        if is_testing():
            def mock_auth_verify_id_token(*args, **kwargs):
                mock_uid = args[0]
                warnings.warn(
                    "@authenticate (decorator for service methods) is returning uid as {}. ".format(
                        mock_uid))
                return {
                    "uid": mock_uid
                }

            return mock_auth_verify_id_token(*args, **kwargs)
        else:
            from firebase_admin import auth
            return auth.verify_id_token(*args, **kwargs)


def default_authentication(id_token) -> (str, int):
    auth_verify_id_token = AuthVerifyIdTokenFunc()
    import firebase_admin.auth
    try:
        # Verify the ID token while checking if the token is revoked by
        # passing check_revoked=True.
        decoded_token = auth_verify_id_token(id_token, check_revoked=True, app=Context.firebase_app)
        # Token is valid and not revoked.
        uid = decoded_token['uid']
        return uid, 200
    except firebase_admin.auth.RevokedIdTokenError:
        # Token revoked, inform the user to re-authenticate or signOut().
        return None, 401
    except firebase_admin.auth.InvalidIdTokenError:
        # Token is invalid
        return None, 402


def authenticate(func):
    """
        Wraps a resource to provide authentication.
        Note that the resource need to take uid in kwargs
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not getattr(func, 'authenticated', True):
            return func(*args, **kwargs)

        if "Authorization" not in request.headers:
            abort(401)

        id_token = request.headers['Authorization'].split(' ').pop()
        uid, status_code = default_authentication(id_token)  # custom account lookup function

        if status_code == 401:
            # 'Unauthorized. Token revoked, inform the user to '
            # 'reauthenticate or signOut(). '
            abort(401)
        elif status_code == 402:
            # 'Invalid token'
            abort(402)

        if uid:
            func_acct = partial(func, uid=uid)
            return func_acct(*args, **kwargs)

        abort(401)

    return wrapper



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="onto",
    # Beta release
    version="0.0.8a5",
    author="Bill Rao",
    author_email="billrao@me.com",
    description="Build reactive back end with ease ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/billyrrr/onto",
    keywords=["firebase", "firestore", "ORM", "flasgger", "flask",
              "backend", "nosql", "reactive", "framework"],
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',

        'Programming Language :: Python :: 3.7',

        'Programming Language :: Python :: 3.8',

        "License :: OSI Approved :: MIT License",

        "Operating System :: OS Independent",
    ],
    install_requires=[
        'Flask>=0.10',
        # TODO: add version constraint
        'marshmallow',
        "inflection",
        "dictdiffer",
        "iso8601",
        "pyyaml",
        "ProxyTypes",
        "pymonad",
        'pytz',
    ],
    extras_require={
        'pony': ['pony'],
        'firestore': [
            'google-auth',
            'google-api-python-client',
            'firebase-admin>=3.0.0',
            'google-cloud-firestore',
            'google-cloud-datastore',
            'grpcio==1.30.0',
        ],
        'gcloud': ["gitpython", "google-cloud-logging",],
        'flink': ['apache-flink'],
        'statefun': ['apache-flink-statefun==3.2.0'],
        'kafka': ["aiokafka"],
        'couchdb': ['couchdb'],
        'couchbase': ['couchbase'],
        'leancloud': ['leancloud'],
        'celery': ['celery'],
        'jsonrpc': ['jsonrpcclient[requests]', 'Flask-JSONRPC'],
        'graphql': ['websockets', 'starlette-graphql-zifulu'],
        'flasgger': ["apispec>=2.0.2",
                     "flasgger"],
        'flask': ['flask-socketio'],
        'rest_api': ['starlette']
    }
    # entry_points = {
    #     'console_scripts': ['`onto`=scripts.deploy:deploy_all'],
    # }
)



class _NA:
    """
    Used when the value of a parameter is not supplied by
        the user. Created in order to differentiate from
        None or False.

    """
    pass


def read_obj_type(d, obj_cls):
    """ Returns obj_type from the raw dictionary

    Note: "obj_type" is NOT the "attribute" property of obj_type field instance
    :param d:
    :param obj_cls:
    :return:
    """
    schema_obj = obj_cls.get_schema_obj()
    if schema_obj is None:
        if "obj_type" in d:
            return d["obj_type"]
        else:
            return None
    else:
        fds = schema_obj.fields
        if "obj_type" in fds:
            return fds["obj_type"].read_obj_type_str(d)
        else:
            return None



"""
ViewModel source:
    Invokes a mediator with instantiated view model

Example: invoke on_create when a form is submitted to a collection
    owned by a user.
"""
from typing import Type

from onto.view_model import ViewModel
from onto.query.query import QueryBase
from onto.source.firestore import FirestoreSource


class ViewModelSource(FirestoreSource):

    def __init__(self, view_model_cls: Type[ViewModel], query: QueryBase):
        self.view_model_cls = view_model_cls
        super().__init__(query=query)

    def _call(self, container):
        for change_type_str, ref, snapshot in self.delta(container):
            obj = self.view_model_cls.from_snapshot(
                ref=ref, snapshot=snapshot)
            self._invoke_mediator(func_name=change_type_str, obj=obj)




from .mediators import CityViewMediator, CityFormMediator
from .models import City
from . import CTX

city_view_mediator = CityViewMediator(
    query=City.get_query()
)

city_form_mediator = CityFormMediator(
    query=CTX.db.collection_group("cityForms")
)

if __name__ == "__main__":
    city_view_mediator.start()
    city_form_mediator.start()



import pytest
from marshmallow import Schema
from onto import fields


def test_basic_schema():

    class BasicSchema(Schema):
        integer_field = fields.Integer()

    class BasicObj:

        def __init__(self):
            self.integer_field = 0

    bs = BasicSchema()
    bo = BasicObj()
    bo.integer_field = 1

    d = bs.dump(bo)
    assert d == {
        "integer_field": 1
    }





from onto.store import SnapshotContainer


def test_set_with_timestamp():
    container = SnapshotContainer()
    container.set_with_timestamp('k', 'v', (1, 0))
    assert container.get('k', (1, 0)) == 'v'


def test_has_previous():
    container = SnapshotContainer()
    assert not container.has_previous('k')
    container.set_with_timestamp('k', 'v', (1, 0))
    assert container.has_previous('k')


def test_previous():
    container = SnapshotContainer()
    container.set_with_timestamp('k', 'v1', (1, 0))
    assert container.previous('k') == 'v1'
    container.set_with_timestamp('k', 'v2', (2, 0))
    assert container.previous('k') == 'v2'


def test_get_with_range():
    container = SnapshotContainer()
    container.set_with_timestamp('k', 'v1', (1, 0))
    container.set_with_timestamp('k', 'v2', (2, 0))
    assert list(container.get_with_range(
        key='k',
        lo_excl=(1, 0),
        hi_incl=(2, 0)
    )) == ['v2']
    assert list(container.get_with_range(
        key='k',
        hi_incl=(2, 0)
    )) == ['v1', 'v2']
